"""Probe mixins containing high-level steps."""

import asyncio
from datetime import datetime, timedelta
import logging
from typing import (
    Callable,
    Iterable,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TYPE_CHECKING,
)

from ya_activity import ExeScriptCommandResult, ExeScriptRequest
from ya_market import AgreementProposal, Demand, DemandOfferBase, Proposal
from ya_payment import Acceptance, Allocation, Invoice

from goth.node import DEFAULT_SUBNET
from goth.runner.step import step

if TYPE_CHECKING:
    from goth.runner.probe import Probe
    from goth.runner.probe.agent import AgentComponent, ProviderAgentComponent
    from goth.runner.probe.rest_client import RestApiComponent

logger = logging.getLogger(__name__)


class ProbeProtocol(Protocol):
    """Protocol class representing the probe interface in mixins.

    This is mainly to fix mypy errors when using `Probe` directly as `self` type.
    """

    agents: "List[AgentComponent]"
    """List of agent components that will be started as part of a probe."""

    api: "RestApiComponent"
    """REST API probe component."""

    name: str
    """Name of the probe container."""


class ActivityApiMixin:
    """Probe mixin providing high-level test steps which use yagna activity API."""

    @step()
    async def create_activity(self: ProbeProtocol, agreement_id: str) -> str:
        """Call create_activity on the activity api."""

        activity_id = await self.api.activity.control.create_activity(agreement_id)
        return activity_id

    @step()
    async def call_exec(self: ProbeProtocol, activity_id: str, exe_script: str) -> str:
        """Call call_exec on the activity api."""

        script_request = ExeScriptRequest(exe_script)
        batch_id = await self.api.activity.control.call_exec(
            activity_id, script_request
        )
        return batch_id

    @step()
    async def collect_results(
        self: ProbeProtocol, activity_id: str, batch_id: str, num_results: int
    ) -> List[ExeScriptCommandResult]:
        """Call collect_results on the activity api."""

        results: List[ExeScriptCommandResult] = []

        while len(results) < num_results:
            results = await self.api.activity.control.get_exec_batch_results(
                activity_id, batch_id
            )
            await asyncio.sleep(1.0)
        return results

    @step()
    async def destroy_activity(self: ProbeProtocol, activity_id: str) -> None:
        """Call destroy_activity on the activity api."""

        await self.api.activity.control.destroy_activity(activity_id)


class MarketApiMixin:
    """Probe mixin providing high-level test steps which use yagna market API."""

    @step()
    async def confirm_agreement(self: ProbeProtocol, agreement_id: str) -> None:
        """Call confirm_agreement on the market api."""
        await self.api.market.confirm_agreement(agreement_id)

    @step()
    async def counter_proposal(
        self: ProbeProtocol,
        subscription_id: str,
        demand: Demand,
        provider_proposal: Proposal,
    ) -> str:
        """Call counter_proposal_demand on the market api."""

        proposal = DemandOfferBase(
            constraints=demand.constraints,
            properties=demand.properties,
        )

        counter_proposal = await self.api.market.counter_proposal_demand(
            subscription_id=subscription_id,
            proposal_id=provider_proposal.proposal_id,
            demand_offer_base=proposal,
        )

        return counter_proposal

    @step()
    async def create_agreement(self: ProbeProtocol, proposal: Proposal) -> str:
        """Call create_agreement on the market api."""

        valid_to = str(datetime.utcnow() + timedelta(days=1)) + "Z"
        logger.debug(
            "Creating agreement, proposal_id=%s, valid_to=%s",
            proposal.proposal_id,
            valid_to,
        )
        agreement_proposal = AgreementProposal(
            proposal_id=proposal.proposal_id, valid_to=valid_to
        )

        agreement_id = await self.api.market.create_agreement(agreement_proposal)
        return agreement_id

    @step()
    async def subscribe_demand(
        self: ProbeProtocol, demand: Demand
    ) -> Tuple[str, Demand]:
        """Call subscribe demand on the market api."""
        subscription_id = await self.api.market.subscribe_demand(demand)
        return subscription_id, demand

    @step()
    async def subscribe_template_demand(
        self: ProbeProtocol, task_package: str, constraints: str
    ) -> Tuple[str, Demand]:
        """Build Demand from template and call subscribe demand on market api."""

        demand = DemandOfferBase(
            properties={
                "golem.node.id.name": "test1",
                "golem.srv.comp.expiration": int(
                    (datetime.now() + timedelta(minutes=10)).timestamp() * 1000
                ),
                "golem.srv.comp.task_package": task_package,
                "golem.node.debug.subnet": DEFAULT_SUBNET,
            },
            constraints=constraints,
        )

        return await self.subscribe_demand(demand)  # type: ignore

    @step()
    async def terminate_agreement(
        self: ProbeProtocol, agreement_id: str, reason: Optional[str]
    ):
        """Call terminate_agreement on the market api."""
        await self.api.market.terminate_agreement(
            agreement_id, request_body={"message": f"Terminated by {self.name}"}
        )

    @step()
    async def unsubscribe_demand(self: ProbeProtocol, subscription_id: str) -> None:
        """Call unsubscribe demand on the market api."""
        await self.api.market.unsubscribe_demand(subscription_id)

    @step()
    async def wait_for_approval(self: ProbeProtocol, agreement_id: str) -> None:
        """Call wait_for_approval on the market api."""
        await self.api.market.wait_for_approval(agreement_id)

    @step()
    async def wait_for_proposals(
        self: ProbeProtocol,
        subscription_id: str,
        providers: Sequence["Probe"],
        filter: Optional[Callable[[Proposal], bool]] = lambda p: True,
    ) -> List[Proposal]:
        """Call collect_offers on the market api.

        Polls collect_offers continously until an offer from each of the given
        providers is received. Returns a list of the collected proposals.
        """
        proposals: List[Proposal] = []
        provider_ids = {p.address for p in providers}

        while len(proposals) < len(provider_ids):
            collected_offers = await self.api.market.collect_offers(subscription_id)
            if collected_offers:
                logger.debug(
                    "collect_offers(%s). collected_offers=%r",
                    subscription_id,
                    collected_offers,
                )
                collected_proposals = [
                    offer.proposal
                    for offer in collected_offers
                    if (
                        offer.proposal.issuer_id in provider_ids
                        and filter(offer.proposal)
                    )
                ]
                proposals.extend(collected_proposals)
            else:
                logger.debug(
                    "Waiting for proposals. subscription_id=%s", subscription_id
                )

        return proposals


class PaymentApiMixin:
    """Probe mixin providing high-level test steps which use yagna payment API."""

    @step()
    async def gather_invoices(self: ProbeProtocol, agreement_id: str) -> List[Invoice]:
        """Call gather_invoice on the payment api."""

        invoices: List[Invoice] = []

        while not invoices:
            await asyncio.sleep(2.0)
            invoices = await self.api.payment.get_invoices()
            invoices = [inv for inv in invoices if inv.agreement_id == agreement_id]

        return invoices

    @step()
    async def pay_invoices(
        self: ProbeProtocol, invoice_events: Iterable[Invoice]
    ) -> None:
        """Call accept_invoice on the payment api."""

        for invoice_event in invoice_events:
            allocation = Allocation(
                allocation_id="",
                total_amount=invoice_event.amount,
                spent_amount=0,
                remaining_amount=0,
                make_deposit=True,
            )
            allocation_result = await self.api.payment.create_allocation(allocation)
            logger.debug("Created allocation. id=%s", allocation_result)

            acceptance = Acceptance(
                total_amount_accepted=invoice_event.amount,
                allocation_id=allocation_result.allocation_id,
            )
            await self.api.payment.accept_invoice(invoice_event.invoice_id, acceptance)
            logger.debug("Accepted invoice. id=%s", invoice_event.invoice_id)


class ProviderProbeProtocol(ProbeProtocol, Protocol):
    """Protocol class representing `ProviderProbe` class.

    This is mainly to fix mypy errors when used as `self` in mixins.
    """

    provider_agent: "ProviderAgentComponent"
    """The agent component running `ya-provider` for this probe.
    This field is added for convenience to make getting this agent instance easier."""


class ProviderLogMixin:
    """Provider probe mixin which adds step functions that wait for specific logs."""

    @step()
    async def wait_for_offer_subscribed(self: ProviderProbeProtocol):
        """Wait until the provider agent subscribes to the offer."""
        await self.provider_agent.wait_for_log("Subscribed offer")

    @step()
    async def wait_for_proposal_accepted(self: ProviderProbeProtocol):
        """Wait until the provider agent subscribes to the offer."""
        await self.provider_agent.wait_for_log("Decided to CounterProposal")

    @step()
    async def wait_for_agreement_approved(self: ProviderProbeProtocol):
        """Wait until the provider agent subscribes to the offer."""
        await self.provider_agent.wait_for_log("Decided to ApproveAgreement")

    @step()
    async def wait_for_exeunit_started(self: ProviderProbeProtocol):
        """Wait until the provider agent starts the exe-unit."""
        await self.provider_agent.wait_for_log(
            r"(.*)\[ExeUnit\](.+)Supervisor initialized$"
        )

    @step()
    async def wait_for_exeunit_finished(self: ProviderProbeProtocol):
        """Wait until exe-unit finishes."""
        await self.provider_agent.wait_for_log(
            "ExeUnit process exited with status Finished - exit code: 0"
        )

    @step()
    async def wait_for_agreement_terminated(self: ProviderProbeProtocol):
        """Wait until Agreement will be terminated.

        This can happen for 2 reasons (both caught by this function):
        - Requestor terminates - most common case
        - Provider terminates - it happens for compatibility with previous
        versions of API without `terminate` endpoint implemented. Moreover
        Provider can terminate, because Agreements condition where broken.
        """
        await self.provider_agent.wait_for_log(r"Agreement \[.*\] terminated by")

    @step()
    async def wait_for_agreement_cleanup(self: ProviderProbeProtocol):
        """Wait until Provider will cleanup all allocated resources.

        This can happen before or after Agreement terminated log will be printed.
        """
        await self.provider_agent.wait_for_log(r"Agreement \[.*\] cleanup finished.")

    @step()
    async def wait_for_invoice_sent(self: ProviderProbeProtocol):
        """Wait until the invoice is sent."""
        await self.provider_agent.wait_for_log("Invoice (.+) sent")

    @step(default_timeout=300)
    async def wait_for_invoice_paid(self: ProviderProbeProtocol):
        """Wait until the invoice is paid."""
        await self.provider_agent.wait_for_log("Invoice .+? for agreement .+? was paid")



"""Transport address helper.

Module containing constants and templates related to addresses commonly used in
the context of yagna and the test harness.

Note: the variables defining ports are also used in the `nginx.conf` file
in the proxy container. If you rename any of them here then make sure to also
apply the same renaming in `nginx.conf`.
"""
from string import Template
from typing import Dict, Mapping, Optional


def ensure_no_trailing_slash(input: str) -> str:
    """Remove the trailing slash from `input`."""
    if input[-1] == "/":
        input = input[:-1]
    return input


class DefaultTemplate(Template):
    """Extend `Template` to allow for default values for template fields."""

    def __init__(self, template: str, default: Dict[str, object]):
        """Set the default and initialise Template()."""

        self.default = default
        super().__init__(template)

    def substitute(self, mapping: Optional[Mapping[str, object]] = None, **kwargs):
        """Replace values in string with `mapping`, merge default and mapping first."""
        return super(DefaultTemplate, self).substitute(
            self._with_default(mapping or {}), **kwargs
        )

    def safe_substitute(self, mapping: Optional[Mapping[str, object]] = None, **kwargs):
        """Replace values in string with `mapping`, merge default and mapping first.

        Safe means that instead of raising a `KeyError` if a value for a placeholder is
        missing, the placeholder will appear in the result string.
        """
        return super(DefaultTemplate, self).safe_substitute(
            self._with_default(mapping or {}), **kwargs
        )

    def __repr__(self):
        return f"<DefaultTemplate template={self.template}, default={self.default},>"

    def __str__(self):
        return self.substitute()

    def _with_default(self, mapping: Mapping[str, object]):
        """Merge `self.default` with `mapping`.

        Make a copy of the defaults mapping and then extend and/or overwrite the copy
        with entries from `mapping`. Allows for passing in a `mapping` dict alongside
        `kwargs` in `substitute` and `safe_substitute`.
        """
        default_copy = self.default.copy()
        default_copy.update(mapping)
        return default_copy


_BASE_URL_TEMPLATE = "$protocol://$host:$port"

PROXY_HOST = "proxy-nginx"

ROUTER_HOST = "router"
ROUTER_PORT = 7477
ROUTER_PROTOCOL = "tcp"
ROUTER_BASE_URL = DefaultTemplate(
    _BASE_URL_TEMPLATE,
    {"host": ROUTER_HOST, "port": ROUTER_PORT, "protocol": ROUTER_PROTOCOL},
)

ACTIVITY_API_URL = Template("$base/activity-api/v1/")
MARKET_API_URL = Template("$base/market-api/v1/")
PAYMENT_API_URL = Template("$base/payment-api/v1/")

YAGNA_BUS_PORT = 6010
YAGNA_BUS_PROTOCOL = "tcp"
YAGNA_BUS_URL = DefaultTemplate(
    _BASE_URL_TEMPLATE, default={"port": YAGNA_BUS_PORT, "protocol": YAGNA_BUS_PROTOCOL}
)

# NOTE: This variable is used in `nginx.conf` file in the proxy container
# also, because it needs to be host-mapped in order for the tests to be runnable
# on non-Linux platforms, it's also independently specified in the definition for the
# `proxy-nginx` service in `docker-compose.yml`
YAGNA_REST_PORT = 6000

YAGNA_REST_PROTOCOL = "http"
YAGNA_REST_URL = DefaultTemplate(
    _BASE_URL_TEMPLATE,
    default={"port": YAGNA_REST_PORT, "protocol": YAGNA_REST_PROTOCOL},
)

# Range of ports on the host which can be mapped to yagna daemons' REST API port
# NOTE: This variable is used in `nginx.conf` file in the proxy container:
HOST_REST_PORT_START = 6001
# NOTE: This variable is used in `nginx.conf` file in the proxy container:
HOST_REST_PORT_END = 6100


# Port used by the mitmproxy instance
# NOTE: This variable is used in `nginx.conf` file in the proxy container:
MITM_PROXY_PORT = 9000

# Port range used by the internal web server spawned by goth for each test
# Ports are rotated and re-used in a circular fashion
WEB_SERVER_PORT_START = 9201
WEB_SERVER_PORT_END = 9251



"""Classes to help configure and create `YagnaContainer`s."""

from pathlib import Path
from typing import Any, ClassVar, Dict, Iterator, Optional, Type, TYPE_CHECKING

from docker import DockerClient
from goth.address import (
    HOST_REST_PORT_END,
    HOST_REST_PORT_START,
    YAGNA_REST_PORT,
)
from goth.runner.container import DockerContainer, DockerContainerConfig
import goth.runner.container.payment as payment
import goth.runner.container.utils as utils
from goth.runner.log import LogConfig

if TYPE_CHECKING:
    from goth.runner.probe import Probe  # noqa: F401

# Default path for mounting assets within a yagna container
ASSET_MOUNT_PATH = Path("/asset")
# Path for mounting payment IDs inside a yagna container
PAYMENT_MOUNT_PATH = ASSET_MOUNT_PATH / "payment"


class YagnaContainerConfig(DockerContainerConfig):
    """Configuration to be used for creating a new `YagnaContainer`."""

    probe_type: Type["Probe"]
    """Python type of the probe to be instantiated from this config."""

    probe_properties: Dict[str, Any]
    """Additional properties to be set on the probe object."""

    environment: Dict[str, str]
    """Environment variables to be set for this container."""

    payment_id: Optional[payment.PaymentId]
    """Custom key and payment accounts to be imported into yagna ID service."""

    def __init__(
        self,
        name: str,
        probe_type: Type["Probe"],
        volumes: Optional[Dict[Path, str]] = None,
        log_config: Optional[LogConfig] = None,
        environment: Optional[Dict[str, str]] = None,
        privileged_mode: bool = False,
        payment_id: Optional[payment.PaymentId] = None,
        **probe_properties,
    ):
        super().__init__(name, volumes or {}, log_config, privileged_mode)
        self.probe_type = probe_type
        self.probe_properties = probe_properties or {}
        self.environment = environment or {}
        self.payment_id = payment_id


class YagnaContainer(DockerContainer):
    """Extension of DockerContainer to be configured for yagna daemons."""

    COMMAND = ["service", "run", "-d", "/"]
    ENTRYPOINT = "/usr/bin/yagna"
    IMAGE = "yagna-goth"

    ports: Dict[int, int] = {}
    """ Port mapping between the Docker host and the container.
        Keys are container port numbers, values are host port numbers. """

    host_port_range: ClassVar[Iterator[int]] = iter(
        range(HOST_REST_PORT_START, HOST_REST_PORT_END)
    )
    """ Keeps track of assigned ports on the Docker host """

    def __init__(
        self,
        client: DockerClient,
        config: YagnaContainerConfig,
        log_config: Optional[LogConfig] = None,
        **kwargs,
    ):
        self.ports = {YAGNA_REST_PORT: YagnaContainer.host_rest_port()}

        super().__init__(
            client=client,
            command=self.COMMAND,
            entrypoint=self.ENTRYPOINT,
            environment=self._prepare_environment(config),
            image=self.IMAGE,
            log_config=log_config,
            name=config.name,
            ports=self.ports,
            volumes=self._prepare_volumes(config),
            privileged=config.privileged_mode,
            **kwargs,
        )

    @classmethod
    def host_rest_port(cls):
        """Return the next host port that can be used for port mapping.

        Raises `OverflowError` if the port to return would exceed the expected range.
        """
        try:
            return next(cls.host_port_range)
        except StopIteration:
            raise OverflowError(f"Port range exceeded. range_end={HOST_REST_PORT_END}")

    def _prepare_volumes(self, config: YagnaContainerConfig) -> Dict[str, dict]:
        volumes_spec = utils.get_volumes_spec(config.volumes)
        if config.payment_id:
            id_volumes = {payment.get_id_directory(): str(PAYMENT_MOUNT_PATH)}
            id_volumes_spec = utils.get_volumes_spec(id_volumes, writable=False)
            volumes_spec.update(id_volumes_spec)
        return volumes_spec

    def _prepare_environment(self, config: YagnaContainerConfig) -> Dict[str, str]:
        env = config.environment
        if config.payment_id:
            accounts_file_name = config.payment_id.accounts_file.name
            mount_path = str(PAYMENT_MOUNT_PATH / accounts_file_name)
            env[payment.ENV_ACCOUNT_LIST] = mount_path
        return env



"""Test harness runner class, creating the nodes and running the scenario."""

import asyncio
from contextlib import asynccontextmanager, AsyncExitStack
from itertools import chain
import logging
import os
from pathlib import Path
import sys
from typing import (
    cast,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
)

import docker

from goth.runner.container.compose import (
    ComposeConfig,
    ComposeNetworkManager,
    run_compose_network,
)
from goth.runner.container.yagna import YagnaContainerConfig
import goth.runner.container.payment as payment
from goth.runner.exceptions import TestFailure, TemporalAssertionError
from goth.runner.log import configure_logging_for_test, LogConfig
from goth.runner.probe import Probe, create_probe, run_probe
from goth.runner.proxy import Proxy, run_proxy
from goth.runner.step import step  # noqa: F401
from goth.runner.web_server import WebServer, run_web_server


logger = logging.getLogger(__name__)

ProbeType = TypeVar("ProbeType", bound=Probe)


class Runner:
    """Manages the nodes and runs the scenario on them."""

    api_assertions_module: Optional[str]
    """Name of the module containing assertions to be loaded into the API monitor."""

    log_dir: Path
    """Directory for all log files created during this test run."""

    test_name: str
    """Name of the test scenario this runner is used in."""

    probes: List[Probe]
    """Probes used for the test run."""

    proxy: Optional[Proxy]
    """An embedded instance of mitmproxy."""

    _test_failure_callback: Callable[[TestFailure], None]
    """A function to be called when `TestFailure` is caught during a test run."""

    _cancellation_callback: Optional[Callable[[], None]]
    """A function to be called when `CancellationError` is caught during a test run.

    If not set, the error is propagated.
    """

    _compose_manager: ComposeNetworkManager
    """Manager for the docker-compose network portion of the test."""

    _exit_stack: AsyncExitStack
    """A stack of `AsyncContextManager` instances to be closed on runner shutdown."""

    _topology: List[YagnaContainerConfig]
    """A list of configuration objects for the containers to be instantiated."""

    _web_server: WebServer
    """A built-in web server."""

    def __init__(
        self,
        base_log_dir: Path,
        compose_config: ComposeConfig,
        test_name: Optional[str] = None,
        api_assertions_module: Optional[str] = None,
        test_failure_callback: Optional[Callable[[TestFailure], None]] = None,
        cancellation_callback: Optional[Callable[[], None]] = None,
        web_root_path: Optional[Path] = None,
        web_server_port: Optional[int] = None,
    ):
        # Set up the logging directory for this runner
        self.test_name = test_name or self._current_pytest_test_name() or ""
        self.log_dir = base_log_dir / self.test_name
        self.log_dir.mkdir(parents=True, exist_ok=True)

        self.api_assertions_module = api_assertions_module
        self.probes = []
        self.proxy = None
        self._exit_stack = AsyncExitStack()
        self._cancellation_callback = cancellation_callback
        self._test_failure_callback = test_failure_callback
        self._compose_manager = ComposeNetworkManager(
            config=compose_config,
            docker_client=docker.from_env(),
        )
        self._web_server = (
            WebServer(web_root_path, web_server_port) if web_root_path else None
        )

    def get_probes(
        self, probe_type: Type[ProbeType], name: str = ""
    ) -> List[ProbeType]:
        """Get probes by name or type.

        `probe_type` can be a type directly inheriting from `Probe`, as well as a
        mixin type used with probes. This type is used in an `isinstance` check.
        """
        probes = self.probes
        if name:
            probes = [p for p in probes if p.name == name]
        probes = [p for p in probes if isinstance(p, probe_type)]
        return cast(List[ProbeType], probes)

    def check_assertion_errors(self) -> None:
        """If any monitor reports an assertion error, raise the first error."""

        probe_agents = chain(*(probe.agents for probe in self.probes))

        monitors = chain.from_iterable(
            (
                (probe.container.logs for probe in self.probes),
                (agent.log_monitor for agent in probe_agents),
                [self.proxy.monitor] if self.proxy else [],
            )
        )
        failed = chain.from_iterable(
            monitor.failed for monitor in monitors if monitor is not None
        )
        for assertion in failed:
            # We assume all failed assertions were already reported
            # in their corresponding log files. Now we only need to raise
            # one of them to break the execution.
            raise TemporalAssertionError(assertion.name)

    def _create_probes(self, scenario_dir: Path) -> None:
        docker_client = docker.from_env()

        for config in self._topology:
            log_config = config.log_config or LogConfig(config.name)
            log_config.base_dir = scenario_dir

            probe = self._exit_stack.enter_context(
                create_probe(self, docker_client, config, log_config)
            )
            self.probes.append(probe)

    def _current_pytest_test_name(self) -> Optional[str]:
        test_name = os.environ.get("PYTEST_CURRENT_TEST")
        if not test_name:
            return None
        logger.debug("Raw current pytest test=%s", test_name)
        # Take only the function name of the currently running test
        test_name = test_name.split("::")[-1].split()[0]
        logger.debug("Cleaned current test dir name=%s", test_name)
        return test_name

    async def _start_nodes(self):
        node_names: Dict[str, str] = {}
        ports: Dict[str, dict] = {}

        # Start the probes' containers and obtain their IP addresses
        for probe in self.probes:
            ip_address = await self._exit_stack.enter_async_context(run_probe(probe))
            node_names[ip_address] = probe.name
            container_ports = probe.container.ports
            ports[ip_address] = container_ports
            logger.debug(
                "Probe for %s started on IP: %s with port mapping: %s",
                probe.name,
                ip_address,
                container_ports,
            )

        node_names[self.host_address] = "Pytest-Requestor-Agent"

        # Stopping the proxy triggers evaluation of assertions at "the end of events".
        # Install a callback to to check for assertion failures after the proxy stops.
        self._exit_stack.callback(self.check_assertion_errors)

        # Start the proxy node. The containers should not make API calls
        # up to this point.
        self.proxy = Proxy(
            node_names=node_names,
            ports=ports,
            assertions_module=self.api_assertions_module,
        )
        await self._exit_stack.enter_async_context(run_proxy(self.proxy))

        # Collect all agent enabled probes and start them in parallel
        awaitables = [probe.start_agents() for probe in self.probes]
        await asyncio.gather(*awaitables)

    @property
    def host_address(self) -> str:
        """Return the host IP address in the docker network used by the containers.

        Both the proxy server and the built-in web server are bound to this address.

        On Mac (and Windows?) there's no network bridge and the services on the host
        don't have access to Docker's internal network. Thus, we need to use a special
        address `host.docker.internal`
        """

        if sys.platform == "linux":
            return self._compose_manager.network_gateway_address
        else:
            return "host.docker.internal"

    @property
    def web_server_port(self) -> Optional[int]:
        """Return the port of the build-in web server."""
        return self._web_server.server_port if self._web_server else None

    @property
    def web_root_path(self) -> Optional[Path]:
        """Return the directory served by the built-in web server."""
        return self._web_server.root_path if self._web_server else None

    @asynccontextmanager
    async def __call__(
        self, topology: List[YagnaContainerConfig]
    ) -> AsyncGenerator["Runner", None]:
        """Set up a test with the given topology and enter the test context.

        This is an async context manager, yielding its `Runner` instance.
        """
        self._topology = topology
        _install_sigint_handler()
        try:
            try:
                await self._enter()
                yield self
            except asyncio.CancelledError:
                if self._cancellation_callback:
                    self._cancellation_callback()
                else:
                    raise
            finally:
                await self._exit()
        except TestFailure as err:
            if self._test_failure_callback:
                self._test_failure_callback(err)
            else:
                logger.info("Runner stopped due to test failure")

    async def _enter(self) -> None:
        self._exit_stack.enter_context(configure_logging_for_test(self.log_dir))
        logger.info("Running test: %s", self.test_name)

        await self._exit_stack.enter_async_context(
            run_compose_network(self._compose_manager, self.log_dir)
        )

        self._create_probes(self.log_dir)

        if self._web_server:
            await self._exit_stack.enter_async_context(
                # listen on all interfaces
                run_web_server(self._web_server, server_address=None)
            )

        await self._start_nodes()

    async def _exit(self):
        logger.info("Test finished: %s", self.test_name)
        await self._exit_stack.aclose()
        payment.clean_up()


def _install_sigint_handler():
    """Install handler that cancels the current task in the current event loop."""
    import signal

    task = asyncio.current_task()
    loop = asyncio.get_event_loop()

    def _sigint_handler(*args):
        logger.warning("Received SIGINT")
        task.cancel()

    loop.add_signal_handler(signal.SIGINT, _sigint_handler)



"""Module responsible for building the yagna Docker image for testing."""

from dataclasses import asdict, dataclass
import logging
import os
from pathlib import Path
import shutil
from tempfile import TemporaryDirectory
from typing import Callable, List, Optional

from goth.project import PROJECT_ROOT
from goth.runner.container.yagna import YagnaContainer
from goth.runner.download import (
    ArtifactDownloader,
    ReleaseDownloader,
    ENV_API_TOKEN,
)
from goth.runner.process import run_command

YAGNA_DOCKERFILE = "yagna-goth.Dockerfile"
YAGNA_DOCKERFILE_DEB = "yagna-goth-deb.Dockerfile"


logger = logging.getLogger(__name__)


EXPECTED_BINARIES = {
    "exe-unit",
    "golemsp",
    "ya-provider",
    "yagna",
}

DEB_RELEASE_REPOS = [
    "ya-service-bus",
    "ya-runtime-wasi",
    "ya-runtime-vm",
]

PROXY_IMAGE = "proxy-nginx"


@dataclass(frozen=True)
class YagnaBuildEnvironment:
    """Configuration for the Docker build process of a yagna image."""

    docker_dir: Path
    """Local path to a directory with Dockerfiles to use for building images."""
    binary_path: Optional[Path] = None
    """Local path to directory or archive with binaries to be included in the image."""
    branch: Optional[str] = None
    """git branch in yagna repo for which to download binaries."""
    commit_hash: Optional[str] = None
    """git commit hash in yagna repo for which to download binaries."""
    deb_path: Optional[Path] = None
    """Local path to .deb file or dir with .deb files to be installed in the image."""
    release_tag: Optional[str] = None
    """Release tag substring used to filter the GitHub release to download."""

    @property
    def is_using_deb(self) -> bool:
        """Return true if this environment is set up to use a .deb yagna release."""
        return not any([self.binary_path, self.branch, self.commit_hash])


async def _build_docker_image(
    image_name: str, dockerfile: Path, setup_context: Callable[[Path], None]
) -> None:
    """Set up a temporary build directory and issue `docker build` command there."""

    with TemporaryDirectory() as temp_path:
        build_dir = Path(temp_path)
        setup_context(build_dir)

        logger.info(
            "Building %s Docker image. dockerfile=%s, build dir=%s",
            image_name,
            dockerfile,
            build_dir,
        )
        command = ["docker", "build", "-t", image_name, str(build_dir)]
        await run_command(command)


async def build_proxy_image(docker_dir: Path) -> None:
    """Build the proxy-nginx Docker image."""

    required_files = (
        Path("goth", "api_monitor", "nginx.conf"),
        Path("goth", "address.py"),
    )
    proxy_dockerfile = docker_dir / f"{PROXY_IMAGE}.Dockerfile"

    def _setup_context(build_dir: Path) -> None:
        nonlocal proxy_dockerfile
        for path in required_files:
            (build_dir / path.parent).mkdir(parents=True, exist_ok=True)
            shutil.copy2(PROJECT_ROOT / path, build_dir / path)
        shutil.copy2(proxy_dockerfile, build_dir / "Dockerfile")

    await _build_docker_image(PROXY_IMAGE, proxy_dockerfile, _setup_context)


async def build_yagna_image(environment: YagnaBuildEnvironment) -> None:
    """Build the yagna Docker image."""

    docker_dir = environment.docker_dir
    dockerfile = docker_dir / (
        YAGNA_DOCKERFILE_DEB if environment.is_using_deb else YAGNA_DOCKERFILE
    )

    await _build_docker_image(
        YagnaContainer.IMAGE,
        dockerfile,
        lambda build_dir: _setup_build_context(build_dir, environment, dockerfile),
    )


def _download_artifact(env: YagnaBuildEnvironment, download_path: Path) -> None:
    downloader = ArtifactDownloader(token=os.environ.get(ENV_API_TOKEN))
    kwargs = {}

    if env.branch:
        kwargs["branch"] = env.branch
    if env.commit_hash:
        kwargs["commit"] = env.commit_hash

    downloader.download(artifact_name="Yagna Linux", output=download_path, **kwargs)


def _download_release(
    download_path: Path, repo: str, tag_substring: str = "", asset_name: str = ""
) -> None:
    downloader = ReleaseDownloader(repo=repo, token=os.environ.get(ENV_API_TOKEN))
    downloader.download(
        output=download_path, asset_name=asset_name, tag_substring=tag_substring
    )


def _find_expected_binaries(root_path: Path) -> List[Path]:
    binary_paths: List[Path] = []

    for root, dirs, files in os.walk(root_path):
        for f in files:
            if f in EXPECTED_BINARIES:
                binary_paths.append(Path(f"{root}/{f}"))

    found = {p.name for p in set(binary_paths)}
    missing = EXPECTED_BINARIES - found

    if len(missing) > 0:
        raise RuntimeError(
            f"Failed to find all binaries required to build a yagna Docker image. "
            f"root_path={root_path}, missing_binaries={missing}"
        )

    return binary_paths


def _setup_build_context(
    context_dir: Path, env: YagnaBuildEnvironment, dockerfile: Path
) -> None:
    """Set up the build context for `docker build` command.

    This function prepares a directory to be used as build context for
    building yagna image. This includes copying the original Dockerfile and creating
    two directories: `bin` and `deb`. Depending on the build environment, these will be
    populated with assets from either the local filesystem or downloaded from GitHub.
    """
    env_dict: dict = asdict(env)
    filtered_env = {k: v for k, v in env_dict.items() if v is not None}
    logger.info(
        "Setting up Docker build context. path=%s, env=%s", context_dir, filtered_env
    )

    context_binary_dir: Path = context_dir / "bin"
    context_deb_dir: Path = context_dir / "deb"
    context_binary_dir.mkdir()
    context_deb_dir.mkdir()

    if env.branch or env.commit_hash:
        _download_artifact(env, context_binary_dir)
    elif env.binary_path:
        if env.binary_path.is_dir():
            logger.info("Using local yagna binaries. path=%s", env.binary_path)
            binary_paths = _find_expected_binaries(env.binary_path)
            logger.debug("Found expected yagna binaries. paths=%s", binary_paths)
            for path in binary_paths:
                shutil.copy2(path, context_binary_dir)
        elif env.binary_path.is_file():
            logger.info("Using local yagna archive. path=%s", env.binary_path)
            shutil.unpack_archive(env.binary_path, extract_dir=str(context_binary_dir))
    else:
        logger.info("Using yagna release. tag_substring=%s", env.release_tag)
        _download_release(context_deb_dir, "yagna", env.release_tag or "", "provider")

    if env.deb_path:
        if env.deb_path.is_dir():
            logger.info("Using local .deb packages. path=%s", env.deb_path)
            shutil.copytree(env.deb_path, context_deb_dir, dirs_exist_ok=True)
        elif env.deb_path.is_file():
            logger.info("Using local .deb package. path=%s", env.deb_path)
            shutil.copy2(env.deb_path, context_deb_dir)
    else:
        for repo in DEB_RELEASE_REPOS:
            _download_release(context_deb_dir, repo)

    logger.debug(
        "Copying Dockerfile. source=%s, destination=%s", dockerfile, context_dir
    )
    shutil.copy2(dockerfile, context_dir / "Dockerfile")
